import 'package:meta/meta.dart';
import 'dfbaserepository.dart';
import '../model/dfdbmodel.dart';
import '../control/dfresponse.dart';
import '../control/dfconverter.dart';

import '../control/dfquery.dart';
import '../control/dfquerycriteria.dart';
import '../field/dfdbfield.dart';
import 'package:sqflite/sqflite.dart';
import '../../security/control/dfsecurity.dart';
import 'package:path/path.dart' as p;


class OpenDbException extends DfError {
  OpenDbException() : super(code: 7100, message: "Fail to open db.");
}

class CloseDbException extends DfError {
  CloseDbException() : super(code: 7101, message: "Fail to close db.");
}

class UpdateDbStructException extends DfError {
  UpdateDbStructException() : super(code: 7102, message: "Fail to update db structure.");
}

class DeleteQueryException extends DfError {
  DeleteQueryException() : super(code: 7103, message: "Fail to execute delete query.");
}

class FetchQueryException extends DfError {
  FetchQueryException() : super(code: 7104, message: "Fail to execute fetch query.");
}

class SaveQueryException extends DfError {
  SaveQueryException() : super(code: 7105, message: "Fail to execute save query.");
}

class DfDbRepository extends DfBaseRepository<DfDbModel> {

  static const DB_UPD_JSON_FIELD = "updated";
  DfDbRepository({@required DfDbModel model}) : super(model: model);

    @override
  Future<List<DfResponse>> delete(List<DfQuery> queries) async {
      List<DfResponse> responses = List<DfResponse>();
      bool dbClose = true;
      Database db;
      Exception globalErr; 
      try {
         db = await _openDb();
         dbClose = false;
          await db.transaction((txn) async {
            for (DfQuery query in queries) {
              try {
                int updCount = 0;
                if (query.criteriaGroups != null) {
                  String criStr = query.criteriaGroups.toString();
                  List criteriaValues = List();
                  query.criteriaGroups.queryGroups.forEach((grp) => grp.criterias.forEach((cri) {
                    criteriaValues.addAll(cri.values);
                  }));

                  //print("Criteria count: ${criteriaValues.length}");
                  updCount = await txn.delete(query.entityJoin.entities[0].entityName,
                    where: criStr, whereArgs: criteriaValues);
                  Map<String,int> updResult = Map<String, int>();
                  updResult[DB_UPD_JSON_FIELD] = updCount;
                  final resp = DfResponse(singleResult: updResult);
                  resp.query = query;
                  responses.add(resp);              
                }
                else {
                  throw DeleteQueryException();
                }
              }
              catch (e,stacktrace) {
                print(stacktrace);
                print(e);
                responses.add(DfResponse(error: DeleteQueryException()));
              }              
            }
          });
          await _closeDb(db);
          dbClose = true;
      }
      on OpenDbException catch(e) { 
        print(e); 
        globalErr = e;
      }
      on CloseDbException catch(e) { 
        print(e);
        globalErr = e;
      }
      /*catch(e) { 
        print(e);
        response.error = DeleteQueryException();
      }*/
      finally {
        if (!dbClose) {
          await _closeDb(db);
          dbClose = true;
        }
      }

      if (globalErr != null) {
        for (DfQuery query in queries) {
          responses.add(DfResponse(error: globalErr));
        }
      }

      return responses;
  }
  
  @override
  Future<DfResponse> fetch(DfQuery query) async {
    DfResponse response = DfResponse();
    bool dbClose = true;
    Database db;
    try {
      String queryStr = "SELECT ";
      String fetchFieldStr = "";

      //print("select field count: ${query.fields.length}");
      query.fields.forEach((keyf,val) {
        DfDbField key = keyf; 
        if (fetchFieldStr.length > 0) {
          fetchFieldStr = "$fetchFieldStr, ";
        }
        fetchFieldStr = "$fetchFieldStr${key.getSelectQueryField()}";
      });

      String queryQues = "";
      if (query.criteriaGroups != null) {
        queryQues = query.criteriaGroups.toString();
        if (queryQues.length > 0)
          queryQues = " WHERE $queryQues";
      }

      String groupByStr = "";
      if (query.groupBy != null) {
        query.groupBy.forEach((itemf) {
          DfDbField item = itemf;
          if (groupByStr.length > 0) {
            groupByStr = "$groupByStr, ";
          }
          groupByStr = "$groupByStr${item.getSelectQueryAlias()}";
        });
        if (groupByStr.length > 0)
          groupByStr = " GROUP BY $groupByStr";
      }

      String orderByStr = "";
      if (query.orderBy != null) {
        query.orderBy.forEach((item) {
          if (orderByStr.length > 0) {
            orderByStr = "$orderByStr, ";
          }
          orderByStr = "$orderByStr${item.toOrderByStr()}";
        });
        if (orderByStr.length > 0)
          orderByStr = " ORDER BY $orderByStr";
      }

      String pagingStr = "";
      if (query.paging != null) {
        int pageNumber = query.paging.pageNumber < 1 ? 1 :  query.paging.pageNumber; 
        int offset = (pageNumber - 1) * query.paging.perPageCount;
        pagingStr = " LIMIT ${query.paging.perPageCount} OFFSET $offset";
      }
      
      queryStr = "$queryStr$fetchFieldStr FROM ${query.entityJoin.toSource()}$queryQues$groupByStr$orderByStr$pagingStr";
      print("Generate query str: $queryStr");

      List criteriaValues = List();
      if (query.criteriaGroups != null) {
        query.criteriaGroups.queryGroups.forEach((grp) => grp.criterias.forEach((cri) {
            if (cri.oper == OperType.like) {
              cri.values.forEach((val) {
                criteriaValues.add("%${(val as String).toLowerCase()}%");
              });
            }
            else {
              criteriaValues.addAll(cri.values);
            }
        }));
      }
      //print("Criteria count: ${criteriaValues.length}");

      Database db = await _openDb();
      dbClose = false;
      response.result = await db.rawQuery(queryStr,criteriaValues);
      await _closeDb(db);
      dbClose = true;
    }
     on OpenDbException catch(e) { 
        print(e); 
        response.error = e; 
      }
      on CloseDbException catch(e) { 
        print(e);
        response.error = e; 
      }
      catch(e,stacktrace) { 
        print(e);
        print(stacktrace);
        response.error = FetchQueryException();
      } 
      finally {
        if (!dbClose) {
          await _closeDb(db);
          dbClose = true;
        }
      }
      response.query = query;

      return response;
    }
  
    @override
    Future<List<DfResponse>> save(List<DfQuery> queries) async {
      //print("Invoke to save() with count: ${queries.length}");
      List<DfResponse> responses = List<DfResponse>();
      bool dbClose = true;
      Database db;
      Exception globalErr;
      try {
        Database db = await _openDb();
        dbClose = false;
        await db.transaction((txn) async {

          for (DfQuery query in queries) {
            try {
              Map<String,dynamic> saveData = Map<String,dynamic>();
              DfError validateError;
              query.fields.forEach((field,val) {
                //print("Loop save for row field: ${field.name} with val $val");
                if (field.fieldType.validator != null) {
                  List<DfError> validateErrors = field.fieldType.validator.validate();
                  if (validateErrors != null && validateErrors.length > 0 && validateError == null) {
                    validateError = validateErrors[0];
                  }
                }
                if (validateError == null) {
                  if (field.fieldType.isDateTime && val is DateTime) {
                    saveData[field.name] = DfConverter.dateToString(val);
                  }
                  else if (field.fieldType.isBoolean && val is bool) {
                    saveData[field.name] = DfConverter.boolToInt(val);
                  }
                  else {
                    saveData[field.name] = val;                    
                  }
                }
              });
              if (validateError != null) {
                print("validateError: ${validateError.message}");
                responses.add(DfResponse(error: validateError));
                continue;
              }

              int updCount = 0;
              if (query.criteriaGroups == null || query.criteriaGroups.toString() == "") {
                String priFieldStr = "";
                List priFieldValues = List();
                List<DfDbField> tblFields = query.entityJoin.entities[0].fields();
                query.fields.forEach((field,val) {
                  
                  bool isPriKey = tblFields.where((f) =>
                    f.primaryKey && f.name.toLowerCase() == field.name.toLowerCase() 
                  ).toList().length > 0;

                  if (isPriKey) {
                    if (priFieldStr.length > 0) {
                      priFieldStr = "$priFieldStr AND ";            
                    }
                    priFieldStr = "$priFieldStr${field.name} = ?";
                    priFieldValues.add(val);
                  }          
                });

                if (priFieldStr.length > 0) {
                  //print("Using update with values: $priFieldValues");
                  updCount = await txn.update(query.entityJoin.entities[0].entityName, saveData,
                  where: priFieldStr, whereArgs: priFieldValues);
                }
              }
              else {
                String criStr = query.criteriaGroups.toString();
                List criteriaValues = List();
                if (query.criteriaGroups != null) {
                  query.criteriaGroups.queryGroups.forEach((grp) {
                    grp.criterias.forEach((cri) {
                      criteriaValues.addAll(cri.values);
                    });
                  });
                }
                //print("Criteria count: ${criteriaValues.length}");
                updCount = await txn.update(query.entityJoin.entities[0].entityName, saveData,
                  where: criStr, whereArgs: criteriaValues);
              }

              if (updCount == 0) {
                updCount = await txn.insert(query.entityJoin.entities[0].entityName, saveData, conflictAlgorithm: ConflictAlgorithm.replace);
              }

              Map<String,int> updResult = Map<String, int>();
              updResult[DB_UPD_JSON_FIELD] = updCount;
              final resp = DfResponse(singleResult: updResult);
              resp.query = query;
              responses.add(resp);
            }
            catch (e,stacktrace) {
              print(e);
              print(stacktrace);
              responses.add(DfResponse(error: SaveQueryException()));
            }                            
          }//for loop query
        });
        await _closeDb(db);
        dbClose = true;
      }
      on OpenDbException catch(e) { 
        print(e); 
        globalErr = e; 
      }
      on CloseDbException catch(e) { 
        print(e);
        globalErr = e; 
      }
      // catch(e) { 
      //   print(e);
      //   response.error = SaveQueryException();
      // }
      finally {
        if (!dbClose) {
          await _closeDb(db);
          dbClose = true;
        }
      }

      if (globalErr != null) {
        for (DfQuery query in queries) {
          responses.add(DfResponse(error: globalErr));
        }
      }

      return responses;
  }

  @override
  Future<void> startInit() async {
    await _updateDbChange();
  }

  Future<String> _getDbFileName() async {
    var databasesPath = await getDatabasesPath();
    return p.join(databasesPath, "${model.dbName}.sqlitedb");
  }

  Future<Database> _openDb() async {
    try {
      String dbFileName = await _getDbFileName();
      if (model.encrypt)
        await DfSecurity.secureOpenFile(dbFileName);
      // open the database
      Database database = await openDatabase(dbFileName, version: 1,
          onCreate: (Database db, int version) async {
        // When creating the db, create the table

        if (model.entities != null) {
          for (var entity in model.entities) {
            await db.execute(entity.toCreateEntityStr());
          }
        }
      });
      return database;
    }
    catch (e,stacktrace) {
      print(stacktrace);
      throw OpenDbException();
    }
  }

  _closeDb(Database db) async {
    try {
      await db.close();
      String dbFileName = await _getDbFileName();
      if (model.encrypt)
        await DfSecurity.secureCloseFile(dbFileName);
    }
    catch (e) {
      print(e);
      throw CloseDbException();
    }
  }

  _updateDbChange() async {
    bool dbClose = true;
    Database db;
    try {
      Database db = await _openDb();
      dbClose = false;
      for (var entity in model.entities) {
        List<Map> list = await db.rawQuery("SELECT sql FROM sqlite_master where tbl_name = '${entity.entityName}'");
        if (list.length == 0) {
          await db.execute(entity.toCreateEntityStr());
        }
        else {
          String sql = list[0]["sql"];
          List<String> existingFields = _extractExistingFields(sql);
          for (var field in entity.fields()) {
            if (!existingFields.contains(field.name.toLowerCase())) {
              await db.execute(entity.genAlterTblAddFieldStr(field));
            }
          }
        }
      }
      await _closeDb(db);
      dbClose = true;
    }
    on OpenDbException catch(e,stacktrace) { 
      print(stacktrace); 
      throw e;
    }
    on CloseDbException catch(e) { 
      print(e);
      throw e; 
    }
    catch(e) { 
      print(e);
      throw UpdateDbStructException(); 
    }
    finally {
      if (!dbClose) {
        await _closeDb(db);
        dbClose = true;
      }
    } 
  }

  List<String> _extractExistingFields(String sql) {
    int firstIdx = sql.indexOf('(');
    int lastIdx = sql.lastIndexOf(')');
    String str = sql.substring(firstIdx+1,lastIdx);
    List<String> parts = str.split(",");
    List<String> result = List<String>();
    parts.forEach((item) {
      List<String> ps = item.trim().split(" ");
      result.add(ps[0].toLowerCase());
    });
    return result;
  }

  @override
  Future<void> stop() async {
  }
}