import 'package:meta/meta.dart';
import 'dfbaserepository.dart';
import '../model/dfsyncmodel.dart';
import '../control/dfresponse.dart';


import '../control/dfquery.dart';
import '../control/dfsyncquery.dart';
import '../control/dfquerycriteria.dart';
import '../control/dfquerycriteriagroups.dart';
import '../control/dfquerycriteriagroup.dart';
import '../control/dfqueryentityjoin.dart';
import '../control/dfqueryorderby.dart';
import '../control/dfconverter.dart';
import 'dfdbrepository.dart';
import 'dfwebrepository.dart';
import '../model/dfdbmodel.dart';
import '../model/dfwebmodel.dart';
import '../field/dfbasefield.dart';
import '../entity/dfbaseentity.dart';
import '../entity/dfwebentity.dart';
import '../invlog/logsetting.dart';
import '../invlog/web/entity/entwebreq.dart';
import '../invlog/web/entity/entwebreqexe.dart';
import '../invlog/web/entity/webreqfield.dart';
import '../invlog/web/entity/webreqexefield.dart';
import 'package:uuid/uuid.dart';
import 'package:nui_web/nui_web.dart';
import 'dart:async';

class LocalStoreException extends DfError {
  LocalStoreException() : super(code: 9101, message: "Fail to cache, so unable to proceed next step: online synchronization.");
}

class DfSyncRepository extends DfBaseRepository<DfSyncModel> {

  DfSyncRepository({@required DfSyncModel model}) : super(model: model);
  DfDbRepository _webFailRepository;
  Timer timer;

  @override
  Future<List<DfResponse>> delete(List<DfQuery> queries) async {
    return await _saveOrDelete(false, queries);
  }

  List<DfQuery> _convertSyncQueryToNativeQueries(DfQuery query) {
    List<DfQuery> result = [];
    DfSyncQuery syncQuery = query as DfSyncQuery;
    result.add( DfQuery(purpose: syncQuery.purpose, queryFields: syncQuery.fields, entityJoin: syncQuery.entityJoin, criteriaGroups: syncQuery.criteriaGroups, orderBy: syncQuery.orderBy, groupBy: syncQuery.groupBy, paging: syncQuery.paging) );
    result.add( DfQuery(purpose: syncQuery.purpose, queryFields: syncQuery.onlineFields, entityJoin: syncQuery.onlineEntityJoin, criteriaGroups: syncQuery.criteriaGroups, orderBy: syncQuery.orderBy, groupBy: syncQuery.groupBy, paging: syncQuery.paging) );
    return result;
  } 
  
  @override
  Future<DfResponse> fetch(DfQuery query) async {
    List<DfQuery> queries = _convertSyncQueryToNativeQueries(query);
    DfQuery dbQuery = queries[0];
    DfQuery webQuery = queries[1];

    DfResponse resp;
    if (model.webRepository != null) {
      resp = await model.webRepository.fetch(webQuery);
    }
    if ((resp == null || resp.error != null) && model.dbRepository != null) {
      DfError webError = resp.error;
      resp = await model.dbRepository.fetch(dbQuery);
      resp.onlineSyncError = webError;
    }
    return resp;
  }
  
  @override
  Future<List<DfResponse>> save(List<DfQuery> queries) async {
    return await _saveOrDelete(true, queries);
  }

  Future<List<DfResponse>> _saveOrDelete(bool isSave, List<DfQuery> queries) async {
    List<DfQuery> dbQueries = [];
    List<DfQuery> webQueries = [];
    queries.forEach((syncQuery) {
      List<DfQuery> queries = _convertSyncQueryToNativeQueries(syncQuery);
      DfQuery dbQuery = queries[0];
      DfQuery webQuery = queries[1];
      dbQueries.add(dbQuery);
      webQueries.add(webQuery);
    });


    List<DfResponse> responses;
    if (model.dbRepository != null) {
      responses = isSave ? await model.dbRepository.save(dbQueries) : await model.dbRepository.delete(dbQueries); 
    }
    bool dbError = false;
    if (responses != null) {
      responses.forEach((resp) {
        if (resp.error != null)
          dbError = true;
      });
    }
    if (dbError) {
      responses.forEach((resp) {
        resp.onlineSyncError = LocalStoreException();
      });
    }
    else if (model.webRepository != null) {
      responses = isSave ? await model.webRepository.save(webQueries) : await model.webRepository.delete(webQueries);
      List<DfBaseEntity> failedWebs = List<DfBaseEntity>();
      List<DfBaseEntity> failedWebStatus = List<DfBaseEntity>();
      for (int i = 0; i < responses.length; i++) {
        DfResponse resp = responses[i];
        DfQuery query = queries[i];

        if (resp.error != null) {
          await _initWebFailRepository();
          resp.onlineSyncError = resp.error;
          Uuid uuid = Uuid();
          resp.onlineSyncFailTrxId = uuid.v4();
          //Save the fail web req to db

          WebReqField webReq = WebReqField.createFieldBasedOnRepo(_webFailRepository);
          webReq.fId.value = resp.onlineSyncFailTrxId;
          DfWebModel webM = model.webRepository.model;
          webReq.fName.value = webM.name;
          webReq.fEncrypt.value = DfConverter.boolToInt(webM.encrypt);
          webReq.fServerUrl.value = webM.serverUrl;
          webReq.fCommonHeaders.value = DfConverter.dictToJsonString(webM.commonHeaders);

          String sslPinningKeys;
          String sslPinningTypeStr;
          if (webM.sslPinningCri != null) {
            sslPinningKeys = DfConverter.arrToString(webM.sslPinningCri.shakeys);
            sslPinningTypeStr = webM.sslPinningCri.shatype.name;
          }

          webReq.fSslPinningList.value = sslPinningKeys;
          webReq.fSslPinningType.value = sslPinningTypeStr;

          String contextPath;
          String pathVariablesStr;
          String reqParamStr;
          String reqBodyStr;
          String headerStr;
          if (query.entityJoin != null && query.entityJoin.entities != null && query.entityJoin.entities.length > 0) {
            DfWebEntity entity = query.entityJoin.entities[0] as DfWebEntity;
            final reqParam = entity.convertToWebReqParam(query);
            if (reqParam != null) {
              pathVariablesStr = DfConverter.arrToString(reqParam.pathVariables);
              reqParamStr = DfConverter.dictToJsonString(reqParam.reqParams);
              reqBodyStr = DfConverter.dictToJsonString(reqParam.reqBody);
            }
            headerStr = DfConverter.dictToJsonString(entity.headers);
            contextPath = entity.entityName;
          }

          webReq.fHeaders.value = headerStr;
          webReq.fContextPath.value = contextPath;
          webReq.fHttpMethod.value = query.purpose.name;
          webReq.fPathVariables.value = pathVariablesStr;
          webReq.fReqParams.value = reqParamStr;
          webReq.fReqBody.value = reqBodyStr;
          webReq.fCreatedAt.value = DfConverter.dateToString(DateTime.now());
          webReq.fSuccess.value = DfConverter.boolToInt(false);
          failedWebs.add(webReq as DfBaseEntity);

          WebReqExeField webReqExe = WebReqExeField.creatFieldBasedOnRepo(_webFailRepository);
          Uuid exeId = Uuid();
          webReqExe.fId.value = exeId.v4();
          webReqExe.fCreatedAt.value = webReq.fCreatedAt.value;
          webReqExe.fWebReqId.value = webReq.fId.value;
          webReqExe.fErrorCode.value = resp.error.code;
          webReqExe.fErrorMessage.value = resp.error.message;
          webReqExe.fResponse.value = DfConverter.arrDictToJsonString(resp.result);
          failedWebStatus.add(webReqExe as DfBaseEntity);
        }//if (resp.error != null)
      }//for

      if (failedWebs.length > 0) {
        List<DfQuery> webReqInsertQueries = List<DfQuery>();
        failedWebs.forEach((failedWeb) {
          DfQuery query = DfQuery(purpose: DfQueryPurpose.ADD, queryFields: failedWeb.toMap(), entityJoin: DfQueryEntityJoin(entities: [WebReq()])  );
          webReqInsertQueries.add(query);
        });
        await _webFailRepository.save(webReqInsertQueries);

        List<DfQuery> webReqExeInsertQueries = List<DfQuery>();
        failedWebStatus.forEach((failedWebStat) {
          DfQuery query = DfQuery(purpose: DfQueryPurpose.ADD, queryFields: failedWebStat.toMap(), entityJoin: DfQueryEntityJoin(entities: [WebReqExe()])  );
          webReqExeInsertQueries.add(query);
        });
        await _webFailRepository.save(webReqExeInsertQueries);
      }
    }//else if (model.webRepository != null) {
    return responses;
  }

  Future<List<DfResponse>> retryFailedWebReq({List<String> onlineSyncFailTrxIds}) async {
    if (! await DfWebRepository.isInternet()) {
      return [DfResponse(error: OfflineException())];
    }
    await _initWebFailRepository();
    WebReq entWebReq = WebReq();
    WebReqField fieldWebReq = WebReqField.createFieldBasedOnRepo(_webFailRepository);
    List<DfBaseField> selFields = entWebReq.fields();

    List<DfQueryCriteria> criterias = [];
    criterias.add(DfQueryCriteria(field: fieldWebReq.fSuccess, oper: OperType.eq, values: [DfConverter.boolToInt(false)] ));
    if (onlineSyncFailTrxIds != null && onlineSyncFailTrxIds.length > 0) {
      criterias.add(DfQueryCriteria(field: fieldWebReq.fId, oper: OperType.inClause, values: onlineSyncFailTrxIds ));
    }
    DfQueryOrderBy orderBy = DfQueryOrderBy(field: fieldWebReq.fCreatedAt, orderByType: OrderByType.asc);

    DfQuery query = DfQuery(purpose: DfQueryPurpose.FETCH, 
                            queryFields: selFields, 
                            entityJoin: DfQueryEntityJoin(entities: [entWebReq]), 
                            criteriaGroups: DfQueryCriteriaGroups(queryGroups: [DfQueryCriteriaGroup(criterias: criterias)]), 
                            orderBy: [orderBy]);

    List<WebReqField> dbWebReqs = List<WebReqField>();

    List<DfBaseEntity> updWebReqSuccess = List<DfBaseEntity>();
    List<DfBaseEntity> insWebReqStatus = List<DfBaseEntity>();

    DfResponse resp = await _webFailRepository.fetch(query);
    if (resp.error != null) {
      print("FetchFailWebReq Error: ${resp.error.message}");
      return [DfResponse(error: resp.error)];
    }
    else {
      List<Map<String,dynamic>> resultSet = resp.result;
      resultSet.forEach((result) {
        DfBaseEntity newWebReq = entWebReq.newInstance();
        newWebReq.fromMap(result);
        WebReqField f = newWebReq as WebReqField;
        dbWebReqs.add(f);
      });

      List<DfResponse> dfResponses = List<DfResponse>();
      for (var dbWebReq in dbWebReqs) {
        NUIWeb nuiweb = await _buildWebRef(dbWebReq.fServerUrl.value, DfConverter.jsonStringToDict(dbWebReq.fCommonHeaders.value));

        String url = "${dbWebReq.fContextPath.value}";
        List<String> pathVariables = DfConverter.stringToList(dbWebReq.fPathVariables.value);
        if (pathVariables != null) {
          pathVariables.forEach((p) {
            if (!url.endsWith("/")) {
              url = "$url/";
            }
            url = "$url$p";
          });
        }

        DfResponse dfResponse;
        switch (dbWebReq.fHttpMethod.value) {
          case DFQUERYPURPOSE_FETCH_NAME :
            NUIHttpResponse nuiResp = await nuiweb.doGET(url: url, params: DfConverter.jsonStringToDict(dbWebReq.fReqParams.value), headers: DfConverter.jsonStringToDict(dbWebReq.fHeaders.value) );
            dfResponse = DfResponse();            
            dfResponse.dataFromHttpResponse(nuiResp);
            break;

          case DFQUERYPURPOSE_ADD_NAME :
            NUIHttpResponse nuiResp = await nuiweb.doPOST(url: url, body: DfConverter.jsonStringToDict(dbWebReq.fReqBody.value), headers: DfConverter.jsonStringToDict(dbWebReq.fHeaders.value) );
            dfResponse = DfResponse();            
            dfResponse.dataFromHttpResponse(nuiResp);
            break;

          case DFQUERYPURPOSE_DELETE :
            NUIHttpResponse nuiResp = await nuiweb.doDELETE(url: url, headers: DfConverter.jsonStringToDict(dbWebReq.fHeaders.value) );
            dfResponse = DfResponse();            
            dfResponse.dataFromHttpResponse(nuiResp);
            break;

          case DFQUERYPURPOSE_UPDATE_WHOLE_ENTITY_ATTRS_NAME :
            NUIHttpResponse nuiResp = await nuiweb.doPUT(url: url, body: DfConverter.jsonStringToDict(dbWebReq.fReqBody.value), headers: DfConverter.jsonStringToDict(dbWebReq.fHeaders.value) );
            dfResponse = DfResponse();            
            dfResponse.dataFromHttpResponse(nuiResp);
            break;

          case DFQUERYPURPOSE_UPDATE_SOME_ENTITY_ATTRS_NAME :
            NUIHttpResponse nuiResp = await nuiweb.doPATCH(url: url, body: DfConverter.jsonStringToDict(dbWebReq.fReqBody.value), headers: DfConverter.jsonStringToDict(dbWebReq.fHeaders.value) );
            dfResponse = DfResponse();            
            dfResponse.dataFromHttpResponse(nuiResp);
            break;          

          default :
            break;
        } //switch
        if (dfResponse != null) {
          dfResponses.add(dfResponse);

          WebReqField webReqObj = WebReqField.createFieldBasedOnRepo(_webFailRepository);
          webReqObj.fId.value = dbWebReq.fId.value;
          webReqObj.fSuccess.value = DfConverter.boolToInt(dfResponse.error == null);
          updWebReqSuccess.add(webReqObj as DfBaseEntity);

          WebReqExeField webReqExeObj = WebReqExeField.creatFieldBasedOnRepo(_webFailRepository);
          Uuid exeId = Uuid();
          webReqExeObj.fId.value = exeId.v4();
          webReqExeObj.fCreatedAt.value = DfConverter.dateToString(DateTime.now());
          webReqExeObj.fWebReqId.value = dbWebReq.fId.value;
          webReqExeObj.fErrorCode.value = resp.error.code;
          webReqExeObj.fErrorMessage.value = resp.error.message;
          webReqExeObj.fResponse.value = DfConverter.arrDictToJsonString(dfResponse.result);
          insWebReqStatus.add(webReqExeObj as DfBaseEntity);

        } //end if (dfResponse != null) 

      } //for loop

      if (updWebReqSuccess.length > 0) {
        List<DfQuery> webReqUpdQueries = List<DfQuery>();
        updWebReqSuccess.forEach((upd) {
          DfQuery query = DfQuery(purpose: DfQueryPurpose.UPDATE_SOME_ENTITY_ATTRS, queryFields: upd.toMap(), entityJoin: DfQueryEntityJoin(entities: [WebReq()])  );
          webReqUpdQueries.add(query);
        });
        await _webFailRepository.save(webReqUpdQueries);

        List<DfQuery> webReqExeInsQueries = List<DfQuery>();
        insWebReqStatus.forEach((ins) {
          DfQuery query = DfQuery(purpose: DfQueryPurpose.ADD, queryFields: ins.toMap(), entityJoin: DfQueryEntityJoin(entities: [WebReqExe()])  );
          webReqExeInsQueries.add(query);
        });
        await _webFailRepository.save(webReqExeInsQueries);
      }
      return dfResponses;
    }//end else - db fetch no error        
  }

  Future<void> _initWebFailRepository() async {
    if (_webFailRepository == null) {
      LogSetting logSetting = LogSetting();
      DfDbModel modelDb = DfDbModel(name: REPO_NAME_SQLITEDB, 
                                  dbName: logSetting.dbName, 
                                  encrypt: logSetting.dataEncrypt, 
                                  entities: [WebReq(),WebReqExe()]);
      _webFailRepository = DfDbRepository(model: modelDb);
      await _webFailRepository.startInit();
    }
  }

  @override
  Future<void> startInit() async {
    _scheduleRetry();
  }  

  @override
  Future<void> stop() async {
    timer?.cancel();
  }

  void _scheduleRetry() {
    if (model.retryIntervalInSecond > 0) {
      timer?.cancel();
      timer = Timer.periodic(Duration(seconds: model.retryIntervalInSecond), (Timer t) {
        if (model.retryIntervalInSecond < 1) {
          timer?.cancel();
        }
        else {
          retryFailedWebReq();
        }
      });
    }      
  }

  Future<NUIWeb> _buildWebRef(String serverUrl, Map<String,String> headers) async {
    var builder = NUIWebBuilder();
    builder.module(Uuid().v4());
    builder.baseURL(serverUrl);
    builder.headers(() => headers);
    return await builder.build();    
  }
}